@@ -54,6 +54,8 @@ typedef enum GlueDataType
 typedef struct GlueTable GlueTable;
 typedef struct GlueTableRow GlueTableRow;
 typedef struct GlueTableDesc GlueTableDesc;
+typedef struct GlueTableRecord GlueTableRecord; 
+
 #endif
 
 // OMG, strings. First of all: casacore::String is a subclass of std::string,
@@ -86,6 +88,7 @@ typedef void (*StringBridgeCallback)(const StringBridge *name, void *ctxt);
 // A more specific callback type for table_get_keyword_info, in which there is
 // additional information we'd like to to transfer.
 typedef void (*KeywordInfoCallback)(const StringBridge *name, GlueDataType dtype, void *ctxt);
+typedef void (*KeywordReprCallback)(const StringBridge *name, GlueDataType dtype, const StringBridge *repr, void *ctxt);
 
 typedef enum TableOpenMode
 {
@@ -134,34 +137,158 @@ extern "C"
 {
     int data_type_get_element_size(const GlueDataType ty);
 
-    GlueTableDesc *tabledesc_create(const StringBridge &type, const TableDescCreateMode mode, 
-                                    ExcInfo &exc);
-    GlueTableDesc *tabledesc_add_scalar_column(GlueTableDesc &table_desc, GlueDataType data_type, 
-                                               const StringBridge &col_name, const StringBridge 
-                                               &comment, bool direct, bool undefined, ExcInfo &exc);
-    GlueTableDesc *tabledesc_add_array_column(GlueTableDesc &table_desc, GlueDataType data_type, 
-                                              const StringBridge &col_name, const StringBridge &comment, 
-                                              bool direct, bool undefined, ExcInfo &exc);
-    GlueTableDesc *tabledesc_add_fixed_array_column(GlueTableDesc &table_desc, GlueDataType data_type, 
-                                                    const StringBridge &col_name, const StringBridge &comment, 
-                                                    const unsigned long n_dims, const unsigned long *dims, 
-                                                    bool direct, bool undefined, ExcInfo &exc);
-    GlueTableDesc *tabledesc_set_ndims(GlueTableDesc &table_desc, const StringBridge &col_name, 
-                                                    const unsigned long n_dims,
-                                                    ExcInfo &exc);
-    GlueTable *table_create(const StringBridge &path, GlueTableDesc &table_desc, 
+    // Table Records
+
+    GlueTableRecord *tablerec_create(ExcInfo &exc);
+    GlueTableRecord *tablerec_copy(const GlueTableRecord &other, ExcInfo &exc);
+    bool tablerec_eq(const GlueTableRecord &rec, const GlueTableRecord &other);
+    int tablerec_get_keyword_info(
+        const GlueTableRecord &rec,
+        KeywordInfoCallback callback,
+        void *ctxt,
+        ExcInfo &exc);
+    int
+    tablerec_get_keyword_repr(
+        const GlueTableRecord &rec, 
+        KeywordReprCallback callback, 
+        void *ctxt, 
+        ExcInfo &exc
+    );
+    int tablerec_get_field_info(
+        const GlueTableRecord &rec,
+        const StringBridge &col_name,
+        GlueDataType *data_type,
+        int *n_dim,
+        unsigned long dims[8],
+        ExcInfo &exc);
+    int tablerec_get_field(
+        const GlueTableRecord &rec,
+        const StringBridge &field_name,
+        void *data,
+        ExcInfo &exc);
+    int tablerec_get_field_string(
+        const GlueTableRecord &rec,
+        const StringBridge &col_name,
+        StringBridgeCallback callback,
+        void *ctxt,
+        ExcInfo &exc);
+    int tablerec_get_field_string_array(
+        const GlueTableRecord &rec,
+        const StringBridge &col_name,
+        StringBridgeCallback callback,
+        void *ctxt,
+        ExcInfo &exc);
+    int tablerec_get_field_subrecord(
+        const GlueTableRecord &rec,
+        const StringBridge &col_name,
+        GlueTableRecord &sub_rec,
+        ExcInfo &exc);
+    int tablerec_put_field(
+        GlueTableRecord &rec,
+        const StringBridge &field_name,
+        const GlueDataType data_type,
+        const unsigned long n_dims,
+        const unsigned long *dims,
+        void *data,
+        ExcInfo &exc);
+    int tablerec_free(GlueTableRecord *rec, ExcInfo &exc);
+
+    // Table Description
+
+    GlueTableDesc *tabledesc_create(
+        const StringBridge &type,
+        const TableDescCreateMode mode,
+        ExcInfo &exc);
+    int tabledesc_add_scalar_column(
+        GlueTableDesc &table_desc,
+        GlueDataType data_type,
+        const StringBridge &col_name,
+        const StringBridge &comment,
+        bool direct,
+        bool undefined,
+        ExcInfo &exc);
+    int tabledesc_add_array_column(
+        GlueTableDesc &table_desc,
+        GlueDataType data_type,
+        const StringBridge &col_name,
+        const StringBridge &comment,
+        bool direct,
+        bool undefined,
+        ExcInfo &exc);
+    int tabledesc_add_fixed_array_column(
+        GlueTableDesc &table_desc,
+        GlueDataType data_type,
+        const StringBridge &col_name,
+        const StringBridge &comment,
+        const unsigned long n_dims,
+        const unsigned long *dims,
+        bool direct,
+        bool undefined,
+        ExcInfo &exc);
+    int tabledesc_set_ndims(
+        GlueTableDesc &table_desc,
+        const StringBridge &col_name,
+        const unsigned long n_dims,
+        ExcInfo &exc);
+    const GlueTableRecord * tabledesc_get_keywords(GlueTableDesc &table_desc, ExcInfo &exc);
+    const GlueTableRecord * tabledesc_get_column_keywords(
+        GlueTableDesc &table_desc, const StringBridge &col_name, ExcInfo &exc);
+    int tabledesc_put_keyword(
+        GlueTableDesc &table_desc,
+        const StringBridge &kw_name,
+        const GlueDataType data_type,
+        const unsigned long n_dims,
+        const unsigned long *dims,
+        void *data,
+        ExcInfo &exc
+    );
+    int tabledesc_put_column_keyword(
+        GlueTableDesc &table_desc,
+        const StringBridge &col_name,
+        const StringBridge &kw_name,
+        const GlueDataType data_type,
+        const unsigned long n_dims,
+        const unsigned long *dims,
+        void *data,
+        ExcInfo &exc
+    );
+    
+    // Table
+
+    GlueTable *table_create(const StringBridge &path, GlueTableDesc &table_desc,
                             unsigned long n_rows, const TableCreateMode mode, ExcInfo &exc);
     GlueTable *table_alloc_and_open(const StringBridge &path, const TableOpenMode mode, ExcInfo &exc);
     void table_close_and_free(GlueTable *table, ExcInfo &exc);
     unsigned long table_n_rows(const GlueTable &table);
     unsigned long table_n_columns(const GlueTable &table);
+    int table_get_file_name(const GlueTable &table, StringBridgeCallback callback, void *ctxt, ExcInfo &exc);
     int table_get_column_names(const GlueTable &table, StringBridgeCallback callback,
                                void *ctxt, ExcInfo &exc);
     unsigned long table_n_keywords(const GlueTable &table);
     int table_get_keyword_info(const GlueTable &table, KeywordInfoCallback callback,
                                void *ctxt, ExcInfo &exc);
-    int table_put_keyword(GlueTable &table, const StringBridge &kw_name, const GlueDataType data_type, 
-                          void *data, ExcInfo &exc);
+    int table_get_column_keyword_info(const GlueTable &table, const StringBridge &col_name,
+                                      KeywordInfoCallback callback, void *ctxt, ExcInfo &exc);
+    const GlueTableRecord *table_get_keywords(GlueTable &table, ExcInfo &exc);
+    const GlueTableRecord *table_get_column_keywords(
+        GlueTable &table,
+        const StringBridge &col_name,
+        ExcInfo &exc);
+    int table_put_keyword(
+        GlueTable &table,
+        const StringBridge &kw_name,
+        const GlueDataType data_type,
+        const unsigned long n_dims,
+        const unsigned long *dims,
+        void *data,
+        ExcInfo &exc);
+    int table_put_column_keyword(
+        GlueTable &table,
+        const StringBridge &col_name,
+        const StringBridge &kw_name,
+        const GlueDataType data_type,
+        const unsigned long n_dims, const unsigned long *dims, void *data,
+        ExcInfo &exc);
     int table_copy_rows(const GlueTable &source, GlueTable &dest, ExcInfo &exc);
     int table_deep_copy_no_rows(const GlueTable &table, const StringBridge &dest_path, ExcInfo &exc);
     int table_get_column_info(const GlueTable &table, const StringBridge &col_name,
@@ -169,12 +296,12 @@ extern "C"
                               int *is_scalar, int *is_fixed_shape, int *n_dim,
                               unsigned long dims[8], ExcInfo &exc);
     int table_remove_column(GlueTable &table, const StringBridge &col_name, ExcInfo &exc);
-    int table_add_scalar_column(GlueTable &table, GlueDataType data_type, const StringBridge &col_name, 
+    int table_add_scalar_column(GlueTable &table, GlueDataType data_type, const StringBridge &col_name,
                                 const StringBridge &comment, bool direct, bool undefined, ExcInfo &exc);
-    int table_add_array_column(GlueTable &table, GlueDataType data_type, const StringBridge &col_name, 
+    int table_add_array_column(GlueTable &table, GlueDataType data_type, const StringBridge &col_name,
                                const StringBridge &comment, bool direct, bool undefined, ExcInfo &exc);
-    int table_add_fixed_array_column(GlueTable &table, GlueDataType data_type, const StringBridge &col_name, 
-                                     const StringBridge &comment, const unsigned long n_dims, 
+    int table_add_fixed_array_column(GlueTable &table, GlueDataType data_type, const StringBridge &col_name,
+                                     const StringBridge &comment, const unsigned long n_dims,
                                      const unsigned long *dims, bool direct, bool undefined, ExcInfo &exc);
     int table_get_scalar_column_data(const GlueTable &table, const StringBridge &col_name,
                                      void *data, ExcInfo &exc);
